# Extract flyby parameters from CASSINI
import zipfile
import os
from datetime import datetime, timedelta

import pandas as pd
import logging
from urllib import request, error
from bs4 import BeautifulSoup

import pydar

## Logging set up for .INFO
logger = logging.getLogger(__name__)
logger.setLevel(logging.INFO)
stream_handler = logging.StreamHandler()
logger.addHandler(stream_handler)

resolution_types = ["B", "D", "F", "H", "I"] # 2, 8, 32, 128, 256 pixels/degree
datafile_types_columns = ["ABDR", "ASUM", "BIDR", "LBDR", "SBDR", "STDR"]

def getFlybyData():
	# Header: Titan flyby id, Radar Data Take Number, Sequence number, Orbit Number/ID
	flyby_id = []
	flyby_radar_take_num = []
	flyby_csv_file = os.path.join(os.path.dirname(__file__), 'data', 'cassini_flyby.csv')  # get file's directory, up one level, /data/*.csv
	flyby_dataframe = pd.read_csv(flyby_csv_file)
	for index, row in flyby_dataframe.iterrows():
		row = row.tolist()
		flyby_id.append(row[0])
		radar_take_num = row[1].split(" ")[1]
		while len(radar_take_num) < 4:
			radar_take_num = "0" + radar_take_num # set all radar take numbers to be four digits long: 229 -> 0229
		flyby_radar_take_num.append(radar_take_num)
	# returns a list of flyby IDs and associated Radar Data Take Number
	return flyby_id, flyby_radar_take_num

def convertFlybyIDToObservationNumber(flyby_id=None):
	# convert Flyby ID to Observation Number to find data files
	pydar.errorHandlingConvertFlybyIDToObservationNumber(flyby_id=flyby_id)

	flyby_csv_file = os.path.join(os.path.dirname(__file__), 'data', 'cassini_flyby.csv')  # get file's directory, up one level, /data/*.csv
	flyby_dataframe = pd.read_csv(flyby_csv_file)
	for index, row in flyby_dataframe.iterrows():
		if row[0] == flyby_id:
			observation_number = row[1].split(" ")[1]
			while len(observation_number) < 4:
				observation_number = "0" + observation_number # set all radar take numbers to be four digits long: 229 -> 0229
			return observation_number

def convertObservationNumberToFlybyID(flyby_observation_num=None):
	# convert Flyby ID to Observation Number to find data files
	pydar.errorHandlingConvertObservationNumberToFlybyID(flyby_observation_num=flyby_observation_num)

	flyby_csv_file = os.path.join(os.path.dirname(__file__), 'data', 'cassini_flyby.csv')  # get file's directory, up one level, /data/*.csv
	flyby_dataframe = pd.read_csv(flyby_csv_file)
	for index, row in flyby_dataframe.iterrows():
		take_ob_num = "0" + row[1].split(" ")[1]
		if take_ob_num == flyby_observation_num:
			return row[0] # returns flyby ID

def retrieveJPLCoradrOptions(flyby_observiation_num):
	# Read JPL Options from CSV
	jpl_coradr_options = []
	coradr_csv_file = os.path.join(os.path.dirname(__file__), 'data', 'coradr_jpl_options.csv')  # get file's directory, up one level, /data/*.csv
	coradr_dataframe = pd.read_csv(coradr_csv_file)
	for index, row in coradr_dataframe.iterrows():
		row = row.tolist()
		jpl_coradr_options.append(row[0])

	find_cordar_listing = 'CORADR_{0}'.format(flyby_observiation_num)
	version_types_avaliable = list(filter(lambda x: find_cordar_listing in x, jpl_coradr_options))
	more_accurate_model_number = version_types_avaliable[-1] # always choose the last and more up to date version number
	logger.info("Most recent CORADR version is {0} from the available list {1}".format(more_accurate_model_number, version_types_avaliable))
	return more_accurate_model_number

def downloadAAREADME(cordar_file_name, segment_id):
	# Download AAREADME.txt within a CORADR directory
	aareadme_name = "AAREADME.TXT"
	aareadme_url = "https://pds-imaging.jpl.nasa.gov/data/cassini/cassini_orbiter/{0}/{1}".format(cordar_file_name, aareadme_name)

	# Retrieve a list of all elements from the base URL to download AAREADME.txt
	logger.info("Retrieving {0} {1}".format(cordar_file_name, aareadme_name))
	aareadme_name = os.path.join("pydar_results/{0}_{1}".format(cordar_file_name, segment_id), aareadme_name)
	try:
		request.urlretrieve(aareadme_url)
	except error.HTTPError as err:
		logger.critical("Unable to access: {0}\nError (and exiting): '{1}'".format(aareadme_url, err.code))
		exit()
	else:
		response = request.urlretrieve(aareadme_url, aareadme_name)

def downloadBIDRCORADRData(cordar_file_name, segment_id, resolution_px):
	# Download BDIR files
	base_url = "https://pds-imaging.jpl.nasa.gov/data/cassini/cassini_orbiter/{0}/DATA/BIDR/".format(cordar_file_name)
	logger.info("Retrieving BIDR filenames from: {0}\n".format(base_url))

	# Retrieve a list of all elements from the base URL to download
	base_html = request.urlopen(base_url).read()
	soup = BeautifulSoup(base_html, 'html.parser')
	table = soup.find('table', {"id": "indexlist"})
	table_text = (table.text).split("\n")
	url_filenames = []
	all_bidr_files = []
	for txt in table_text:
		if txt.startswith('BI'):
			filename = (txt.split('/')[0]).split(".")[0]
			if 'LBL' in (txt.split('/')[0]).split(".")[1]:
				filename += '.LBL'
			if 'ZIP' in (txt.split('/')[0]).split(".")[1]:
				filename += '.ZIP'
			all_bidr_files.append(filename)
			if segment_id in filename: # only save certain segements
				for resolution in resolution_px: # only save top x resolutions
					bi_types = ["B", "E", "T", "N", "M", "L"] # BI<OPTION>Q<RESOLUTION>
					for bi in bi_types:
						if "BI{0}Q{1}".format(bi, resolution) in filename:
							url_filenames.append(filename)

	logger.info("All BIDR files found with specified resolution, segment, and flyby identification: {0}\n".format(url_filenames))
	if len(url_filenames) == 0:
		logger.critical("No BIDR files found with resolution, segment, and flyby identification. Please use different parameters to retrieve data.\nAll files found: {0}".format(all_bidr_files))
		exit()

	for i, coradr_file in enumerate(url_filenames):
		if 'LBL' in coradr_file:
			label_url = "https://pds-imaging.jpl.nasa.gov/data/cassini/cassini_orbiter/{0}/DATA/BIDR/{1}".format(cordar_file_name, coradr_file)
			logger.info("Retrieving [{0}/{1}]: {2}".format(i+1, len(url_filenames), label_url))
			label_name = label_url.split("/")[-1].split(".")[0] + ".LBL"
			label_name = os.path.join("pydar_results/{0}_{1}".format(cordar_file_name, segment_id), label_name)
			try:
				request.urlretrieve(label_url)
			except error.HTTPError as err:
				logger.critical("Unable to access: {0}\nError (and exiting): '{1}'".format(label_url, err.code))
				exit()
			else:
				response = request.urlretrieve(label_url, label_name)
		if 'ZIP' in coradr_file:
			data_url = "https://pds-imaging.jpl.nasa.gov/data/cassini/cassini_orbiter/{0}/DATA/BIDR/{1}".format(cordar_file_name, coradr_file)
			logger.info("Retrieving [{0}/{1}]: {2}".format(i+1, len(url_filenames), data_url))
			zipfile_name = data_url.split("/")[-1].split(".")[0] + ".zip"
			zipfile_name = os.path.join("pydar_results/{0}_{1}".format(cordar_file_name, segment_id), zipfile_name)
			try:
				request.urlretrieve(data_url)
			except error.HTTPError as err:
				logger.info("Unable to access: {0}\nError (and exiting): '{1}'".format(data_url, err.code))
				exit()
			else:
				response = request.urlretrieve(data_url, zipfile_name)
				zipped_image = zipfile_name.split(".")[0] + ".IMG"
				with zipfile.ZipFile(zipfile_name, 'r') as zip_ref:
					zipped_image_path = os.path.join("pydar_results/{0}_{1}".format(cordar_file_name, segment_id))
					zip_ref.extractall(zipped_image_path)

def downloadSBDRCORADRData(cordar_file_name, segment_id):
	# Download SBDR files
	base_url = "https://pds-imaging.jpl.nasa.gov/data/cassini/cassini_orbiter/{0}/DATA/SBDR/".format(cordar_file_name)
	logger.info("\nRetrieving SBDR filenames from: {0}".format(base_url))

	# Retrieve a SBDR file from filename at SBDR URL
	base_html = request.urlopen(base_url).read()
	soup = BeautifulSoup(base_html, 'html.parser')
	table = soup.find('table', {"id": "indexlist"})
	table_text = (table.text).split("\n")
	sbdr_files = []
	sbdr_filename = ''
	for txt in table_text:
		if txt.startswith('SBDR'):
			sbdr_filename = (txt.split('/')[0]).split(".")[0]
			if 'TAB' in (txt.split('/')[0]).split(".")[1]: # TAB information
				sbdr_filename += '.TAB'
			if 'FMT' in (txt.split('/')[0]).split(".")[1]: # FMT information required to read TAB
				sbdr_filename += '.FMT'
			sbdr_files.append(sbdr_filename)

	logger.info("SBDR files found: {0}".format(sbdr_files))
	if len(sbdr_files) == 0:
		logger.critical("No SBDR files were found with resolution, segment, and flyby identification. Please use different parameters to retrieve data")
		exit()

	for sbdr_file in sbdr_files:
		sbdr_url = "https://pds-imaging.jpl.nasa.gov/data/cassini/cassini_orbiter/{0}/DATA/SBDR/{1}".format(cordar_file_name, sbdr_file)
		logger.info("Retrieving SBDR file '{0}': {1}".format(sbdr_file, sbdr_url))
		sbdr_name = os.path.join("pydar_results/{0}_{1}".format(cordar_file_name, segment_id), sbdr_file)
		try:
			request.urlretrieve(sbdr_url)
		except error.HTTPError as err:
			logger.critical("Unable to access: {0}\nError (and exiting): '{1}'".format(sbdr_url, err.code))
			exit()
		else:
			response = request.urlretrieve(sbdr_url, sbdr_name)

def downloadAdditionalDataTypes(cordar_file_name, segment_id, additional_data_type):
	# Download additional data types
	additional_data_url = "https://pds-imaging.jpl.nasa.gov/data/cassini/cassini_orbiter/{0}/DATA/{1}".format(cordar_file_name, additional_data_type)
	logger.info("\n[TODO: does not currently download] '{0}': {1}".format(additional_data_type, additional_data_url))
	# TODO: add functionality for which files should be downloaded
	# This function does not currently have functionality in pydar

def extractFlybyDataImages(flyby_observation_num=None,
							flyby_id=None,
							segment_num=None,
							additional_data_types_to_download=[],
							resolution='I',
							top_x_resolutions=None):

	if flyby_id is not None and type(flyby_id) == str:
		flyby_id = flyby_id.upper() # ensure that observation number set to capitalized 'T'
	if flyby_observation_num is not None and type(flyby_observation_num) == str:
		while len(flyby_observation_num) < 4:
			flyby_observation_num = "0" + flyby_observation_num # set all radar take numbers to be four digits long: 229 -> 0229
	if top_x_resolutions is not None:
		resolution = None # set default resolution to None if selecting the top x resolutions

	# Error handling:
	pydar.errorHandlingExtractFlybyDataImages(flyby_observation_num=flyby_observation_num,
											flyby_id=flyby_id,
											segment_num=segment_num,
											additional_data_types_to_download=additional_data_types_to_download,
											resolution=resolution,
											top_x_resolutions=top_x_resolutions)

	logger.info("flyby_observation_num = {0}".format(flyby_observation_num))
	logger.info("flyby_id = {0}".format(flyby_id))
	logger.info("segment_num = {0}".format(segment_num))
	logger.info("additional_data_types_to_download = {0}".format(additional_data_types_to_download))
	logger.info("resolution = {0}".format(resolution))
	logger.info("top_x_resolutions = {0}".format(top_x_resolutions))

	download_files = True # for debugging, does not always download files before running data

	# update csv
	days_between_checking_jpl_website = 7 # set to 0 to re-run currently without waiting
	x_days_ago = datetime.now() - timedelta(days=days_between_checking_jpl_website)
	filetime = datetime.fromtimestamp(os.path.getctime(os.path.join(os.path.dirname(__file__), 'data', 'coradr_jpl_options.csv')))
	if filetime < x_days_ago:
		# File it more than X days old
		logger.info("file is older than {0} days, running html capture to update coradr_jpl_options.csv (will take about twenty minutes):".format(days_between_checking_jpl_website))
		pydar.csvCORADRJPLOptions() # coradr_jpl_options.csv
		pydar.csvSwathCoverage() # swath_coverage_by_time_position.csv
		pydar.csvFeatureNameDetails() #feature_name_details.csv

	if flyby_id is not None:  # convert flyby Id to an Observation Number
		flyby_observation_num = convertFlybyIDToObservationNumber(flyby_id)

	avaliable_flyby_id, avaliable_observation_numbers = getFlybyData()

	if flyby_observation_num not in avaliable_observation_numbers:
		logger.critical("Observation number '{0}' NOT FOUND in available observation numbers: {1}\n".format(flyby_observation_num, avaliable_observation_numbers))
		exit()
	else:
		logger.debug("Observation number '{0}' FOUND in available observation numbers: {1}\n".format(flyby_observation_num, avaliable_observation_numbers))

	# Download information from pds-imaging site for CORADR
	flyby_observation_cordar_name = retrieveJPLCoradrOptions(flyby_observation_num)
	if not os.path.exists('pydar_results'): os.makedirs('pydar_results')
	if not os.path.exists("pydar_results/{0}_{1}".format(flyby_observation_cordar_name, segment_num)): os.makedirs("pydar_results/{0}_{1}".format(flyby_observation_cordar_name, segment_num))

	if download_files:
		# AAREADME.TXT
		downloadAAREADME(flyby_observation_cordar_name, segment_num)
		
		# BIDR
		if top_x_resolutions is not None:
			downloadBIDRCORADRData(flyby_observation_cordar_name, segment_num, resolution_types[-top_x_resolutions:])
		else:
			downloadBIDRCORADRData(flyby_observation_cordar_name, segment_num, resolution)

		# SBDR
		downloadSBDRCORADRData(flyby_observation_cordar_name, segment_num)

		# Download additional data types
		for data_type in additional_data_types_to_download:
			if data_type not in ["BIDR", "SBDR"]: # ignore data files that have already been downloaded
				downloadAdditionalDataTypes(flyby_observation_cordar_name, segment_num, data_type)

	if len(os.listdir("pydar_results/{0}_{1}".format(flyby_observation_cordar_name, segment_num))) == 0:
		logger.critical("Unable to find any images with current parameters")
		exit()



from .pydar_testing import testing

# extract_flyby_parameters.py function calls
from .extract_flyby_parameters import extractFlybyDataImages
from .extract_flyby_parameters import getFlybyData
from .extract_flyby_parameters import convertFlybyIDToObservationNumber
from .extract_flyby_parameters import convertObservationNumberToFlybyID

# extract_flyby_parameters.py data
from .extract_flyby_parameters import resolution_types
from .extract_flyby_parameters import datafile_types_columns

# display_image.py function calls
from .display_image import displayImages

# retrieve_supplementary_backplanes.py function calls
from .retrieve_supplementary_backplanes import extractMetadata

# read_readme.py function calls
from .read_readme import returnAllAAREADMEOptions
from .read_readme import readAAREADME
from .read_readme import returnAllLBLOptions
from .read_readme import readLBLREADME

# update_csv.py function calls
from .update_csv import csvCORADRJPLOptions
from .update_csv import csvSwathCoverage
from .update_csv import csvFeatureNameDetails

# retrieve_ids_by_time_position.py function calls
from .retrieve_ids_by_time_position import retrieveIDSByFeatureName
from .retrieve_ids_by_time_position import retrieveIDSByLatitudeLongitude
from .retrieve_ids_by_time_position import retrieveIDSByLatitudeLongitudeRange
from .retrieve_ids_by_time_position import retrieveIDSByTime
from .retrieve_ids_by_time_position import retrieveFeaturesFromLatitudeLongitude
from .retrieve_ids_by_time_position import retrieveFeaturesFromLatitudeLongitudeRange

# error_handling.py function calls for testing
from .error_handling import errorHandlingExtractFlybyDataImages
from .error_handling import errorHandlingConvertFlybyIDToObservationNumber
from .error_handling import errorHandlingConvertObservationNumberToFlybyID
from .error_handling import errorHandlingDisplayImages
from .error_handling import errorHandlingREADME
from .error_handling import errorHandlingRetrieveIDSByFeature
from .error_handling import errorHandlingRetrieveByLatitudeLongitude
from .error_handling import errorHandlingRetrieveByLatitudeLongitudeRange
from .error_handling import errorHandlingRetrieveIDSByTime



# Update CSVs

import logging
import os
import re

from bs4 import BeautifulSoup
import pandas as pd
from urllib import request, error

import pydar

## Logging set up for .INFO
logger = logging.getLogger(__name__)
logger.setLevel(logging.INFO)
stream_handler = logging.StreamHandler()
logger.addHandler(stream_handler)

def csvCORADRJPLOptions():
	# runs to access the most up to date optiosn from the JPL webpage
	# Generate: coradr_jpl_options.csv

	# BeautifulSoup web scrapping to find observation file number full title
	logger.info("Refreshing: coradr_jpl_options.csv")
	logger.info("Retrieving observation information from pds-imaging.jpl.nasa.gov/ata/cassini/cassini_orbital....")
	cassini_root_url = "https://pds-imaging.jpl.nasa.gov/data/cassini/cassini_orbiter"
	cassini_html = request.urlopen(cassini_root_url).read()
	soup = BeautifulSoup(cassini_html, 'html.parser')
	table = soup.find('table', {"id": "indexlist"})
	table_text = (table.text).split("\n")
	coradr_options = []
	for txt in table_text:
		if 'CORADR' in txt:
			coradr_title = txt.split('/')[0]
			if '.' not in coradr_title:
				coradr_options.append([coradr_title, False, False, False, False, False, False, False])
		_, flyby_radar_take_num = pydar.getFlybyData()

	# Check of CORADR has specific data files formats
	for i, coradr_id in enumerate(coradr_options):
		coradr_url = "{0}/{1}/DATA".format(cassini_root_url, coradr_id[0])
		logger.info("Retrieving data types [{0}/{1}]: {2}".format(i+1, len(coradr_options), coradr_url))
		coradr_html = request.urlopen(coradr_url).read()
		soup = BeautifulSoup(coradr_html, 'html.parser')
		table = soup.find('table', {"id": "indexlist"})
		table_text = (table.text).split("\n")
		for txt in table_text:
			for i, data_type in enumerate(pydar.datafile_types_columns):
				if data_type in txt:
					coradr_id[i+2] = True
				if coradr_id[0].split("_")[1] in flyby_radar_take_num:
					coradr_id[1] = True # Is a Titan Flyby

	# Wrte to CSV
	header_options = ["CORADR ID",
					"Is a Titan Flyby",
					"Contains ABDR",
					"Contains ASUM",
					"Contains BIDR",
					"Contains LBDR",
					"Contains SBDR",
					"Contains STDR"]
	df = pd.DataFrame(coradr_options, columns=header_options)
	df = df.sort_values(by=["CORADR ID"])
	df.to_csv(os.path.join(os.path.dirname(__file__), 'data', 'coradr_jpl_options.csv'), header=header_options, index=False)

def csvSwathCoverage():
	# generate swath_coverage_by_time_position.csv
	# Estimated runtime: 15 minutes
	logger.info("Refreshing: swath_coverage_by_time_position.csv")

	# Get all Titan Flybys with most up to date versions
	coradr_ids = []
	ids_with_no_bidr = []
	coradr_csv_file = os.path.join(os.path.dirname(__file__), 'data', 'coradr_jpl_options.csv')  # get file's directory, up one level, /data/*.csv
	coradr_dataframe = pd.read_csv(coradr_csv_file)
	for index, row in coradr_dataframe.iterrows():
		row = row.tolist()
		if row[1] == True:
			if row[4] == False: # if is a Titan flyby but "Contains BIDR" is False
				ids_with_no_bidr.append(row[0])
			if coradr_ids != []:
				if coradr_ids[-1].split("_V")[0] in row[0]: # replace the older version with the most recent version
					coradr_ids[-1] = row[0]
				else:
					coradr_ids.append(row[0])
			else:
				coradr_ids.append(row[0])

	# Retrieve a list of all the .lbl for each CORADR ID (different for each resolution)
	data_type_dict = {"F":"Primary Dataset (Linear Scale)", 
					"B": "Primary Dataset in Unsigned Byte Format (Normalized dB)",
					"S": "Normalized SAR (Physical Scale) with Thermal/Quantized Noise Removed",
					"X": "Noise for SAR without Incidence Angle Correction (Physical Scale)",
					"U":"Sigma0 without Incidence Angle (Linear Scale)", 
					"D": "Subtracted STD SAR",
					"E":"Incidence Angle Map", 
					"T":"Latitude Map",
					"N":"Longitude Map",
					"M":"Beam Mask Map",
					"L":"Number of Looks Map"}
	resolution_dict = {"B":2, "D":8, "F":32, "G": 64, "H":128, "I":256} #  pixels/degree
	lbl_information = []
	for radar_id in coradr_ids:
		if radar_id not in ids_with_no_bidr:
			base_url = "https://pds-imaging.jpl.nasa.gov/data/cassini/cassini_orbiter/{0}/DATA/BIDR/".format(radar_id)
			base_html = request.urlopen(base_url).read()
			soup = BeautifulSoup(base_html, 'html.parser')
			table = soup.find('table', {"id": "indexlist"})
			table_text = (table.text).split("\n")
			url_filenames = []
			all_bidr_files = []
			for txt in table_text:
				if txt.startswith('BI'):
					filename = (txt.split('/')[0]).split(".")[0]
					if 'LBL' in (txt.split('/')[0]).split(".")[1]:
						lbl = []
						filename += '.LBL'
						bidr_url = "https://pds-imaging.jpl.nasa.gov/data/cassini/cassini_orbiter/{0}/DATA/BIDR/{1}".format(radar_id, filename)
						logger.info("Retrieving LBL information: {0}".format(bidr_url))
						lbl = [radar_id, None, None, None, None, None, None, None, None, None, None, None, None, None]
						lbl[1] = pydar.convertObservationNumberToFlybyID(radar_id.split("_")[1])
						lbl[2] = (filename.split("_")[2]).split("S")[1] # Segment Number
						lbl[3] = filename
						lbl[4] = filename[2] # Data Type
						lbl[5] = data_type_dict[filename[2]] # Data Type, with full title
						lbl[6] = resolution_dict[filename[4]] # Resolution
						with request.urlopen(bidr_url) as lbl_file:
							for line in (lbl_file.read().decode("UTF-8")).split("\n"):
								if "TARGET_NAME" in line:
									lbl[7] = line.split("=")[1].strip()
								if "MAXIMUM_LATITUDE" in line:
									max_lat = line.split("=")[1].strip()
									max_lat = max_lat.split("<")[0] 
									lbl[8] = max_lat
								if "MINIMUM_LATITUDE" in line:
									min_lat = line.split("=")[1].strip()
									min_lat = min_lat.split("<")[0] 
									lbl[9] = min_lat
								if "EASTERNMOST_LONGITUDE" in line:
									east_long = line.split("=")[1].strip()
									east_long = east_long.split("<")[0] 
									lbl[10] = east_long
								if "WESTERNMOST_LONGITUDE" in line:
									west_long = line.split("=")[1].strip()
									west_long = west_long.split("<")[0] 
									lbl[11] = west_long
								if "START_TIME" in line:
									lbl[12] = line.split("=")[1].strip()
								if "STOP_TIME" in line:
									lbl[13] = line.split("=")[1].strip()
							lbl_information.append(lbl)

	# Wrte to CSV
	header_options = ["CORADR ID",
					"FLYBY ID",
					"SEGMENT NUMBER",
					"FILENAME",
					"DATE TYPE SYMBOL",
					"DATE TYPE",
					"RESOLUTION (pixels/degrees)",
					"TARGET_NAME",
					"MAXIMUM_LATITUDE (Degrees)",
					"MINIMUM_LATITUDE (Degrees)",
					"EASTERNMOST_LONGITUDE (Degrees)",
					"WESTERNMOST_LONGITUDE (Degrees)",
					"START_TIME",
					"STOP_TIME"
					]
	df = pd.DataFrame(lbl_information, columns=header_options)
	df = df.sort_values(by=["CORADR ID"])
	df.to_csv(os.path.join(os.path.dirname(__file__), 'data', 'swath_coverage_by_time_position.csv'), header=header_options, index=False)

def csvFeatureNameDetails():
	# runs to access the most up to date optiosn from the JPL webpage
	# Generate: feature_name_details.csv
	# Estimated runtime: 3 minutes

	# BeautifulSoup web scrapping to find observation file number full title
	logger.info("Refreshing: feature_name_details.csv")
	logger.info("Retrieving observation information from https://planetarynames.wr.usgs.gov/SearchResults?Target=74_Titan....")
	titan_root_url = "https://planetarynames.wr.usgs.gov/SearchResults?Target=74_Titan"
	titan_html = request.urlopen(titan_root_url).read()
	soup = BeautifulSoup(titan_html, 'html.parser')
	#table_feature_names = soup.find_all("td", {"class":"featureNameColumn"})
	#feature_name_lst = [feature_name.text.strip() for feature_name in table_feature_names]
	ahref_feature_names = soup.findAll('a')
	ahref_lst = []
	for link in ahref_feature_names:
		feature_link = link.get('href')
		if feature_link is not None:
			if feature_link.startswith("/Feature/"):
				if feature_link != "/Feature/7014": # ignore a dropped column for 'Sotra Facula'
					ahref_lst.append(feature_link)

	feature_options = []
	base_url = "https://planetarynames.wr.usgs.gov"
	for i, feature_ahref in enumerate(ahref_lst):
		feature_html = request.urlopen(base_url + feature_ahref)
		logger.info("[{0}/{1}] Retrieving: {2}".format(i+1, len(ahref_lst), base_url + feature_ahref))
		soup = BeautifulSoup(feature_html, 'html.parser')
		table = soup.find("div", {"id":"layout_content_wrapper"})
		tr = table.find_all("tr")
		feature_object = [None, None, None, None, None, None, None, None]
		for i in tr:
			feature_row = ((i.text).lstrip()).split("\n")
			feature_row = [f.strip() for f in feature_row if f != '' and re.search('[a-zA-Z-?\d+]', f)]
			if len(feature_row) == 2:
				if feature_row[0] == "Feature Name":
					feature_object[0] = feature_row[1]
				if feature_row[0] == "Northernmost Latitude":
					feature_object[1] = feature_row[1].split(" ")[0]
				if feature_row[0] == "Southernmost Latitude":
					feature_object[2] = feature_row[1].split(" ")[0]
				if feature_row[0] == "Easternmost Longitude":
					feature_object[3] = feature_row[1].split(" ")[0]
				if feature_row[0] == "Westernmost Longitude":
					feature_object[4] = feature_row[1].split(" ")[0]
				if feature_row[0] == "Center Latitude":
					feature_object[5] = feature_row[1].split(" ")[0]
				if feature_row[0] == "Center Longitude":
					feature_object[6] = feature_row[1].split(" ")[0]
				if feature_row[0] == "Origin":
					feature_object[7] = feature_row[1]
		feature_options.append(feature_object)

	huygens_landing_site = ["Huygens Landing Site", 
							"-10.576",
							"-10.576",
							"167.547",
							"167.547",
							"-10.576",
							"167.547",
							"where the Huygens probe landed east Adiri"]
	feature_options.append(huygens_landing_site) # Add Huygens landing site manually

	# Wrte to CSV
	header_options = ['Feature Name', 
					'Northernmost Latitude',
					'Southernmost Latitude',
					'Easternmost Longitude',
					'Westernmost Longitude',
					'Center Latitude',
					'Center Longitude', 
					'Origin of Name']
	df = pd.DataFrame(feature_options, columns=header_options)
	df = df.sort_values(by=["Feature Name"])
	df.to_csv(os.path.join(os.path.dirname(__file__), 'data', 'feature_name_details.csv'), header=header_options, index=False)



## Display Image
import logging
import os

from planetaryimage import PDS3Image
import matplotlib.pyplot as plt

import pydar

## Logging set up for .INFO
logger = logging.getLogger(__name__)
logger.setLevel(logging.INFO)
stream_handler = logging.StreamHandler()
logger.addHandler(stream_handler)

def displayImages(image_directory=None):
	# Display all images in the image directory specified
	pydar.errorHandlingDisplayImages(image_directory=image_directory)
	for filename in os.listdir(image_directory):
		if 'IMG' in filename:
			image_file = os.path.join("{0}/{1}".format(image_directory, filename))
			logger.info("Displaying Image: {0}".format(image_file))
			image = PDS3Image.open(image_file)
			logger.info("Displaying Dimensions: {0}".format(image.shape))
			fig = plt.figure(figsize=(4,6), dpi=120)
			plt.title(filename)
			plt.xlabel("Pixels #")
			plt.ylabel("Pixels #")
			plt.imshow(image.image, cmap='gray')
			plt.show()



########################################################################
# ERROR CATCHES AND LOGGING
########################################################################
import logging
import os
import csv

import pandas as pd

import pydar

## Logging set up for .INFO
logger = logging.getLogger(__name__)
logger.setLevel(logging.CRITICAL)
stream_handler = logging.StreamHandler()
logger.addHandler(stream_handler)

def errorHandlingExtractFlybyDataImages(flyby_observation_num=None,
										flyby_id=None,
										segment_num=None,
										additional_data_types_to_download=[],
										resolution=None,
										top_x_resolutions=None):
	# Error Handling for extract_flyby_parameters variables
	avaliable_flyby_id, avaliable_observation_numbers = pydar.getFlybyData()

	if flyby_observation_num is None and flyby_id is None:
		logger.critical("\nCRITICAL ERROR: Requires either a flyby_observation_num OR flyby_id.\nAvaliable flyby_observation_num: {0}\nAvaliable flyby_id: {1}".format(avaliable_flyby_id, avaliable_observation_numbers))
		exit()

	if flyby_id is not None:
		if type(flyby_id) != str:
			logger.critical("\nCRITICAL ERROR, [flyby_observation_num]: Must be a str, current type = '{0}'".format(type(flyby_id)))
			exit()
		if flyby_id not in avaliable_flyby_id:
			logger.critical("\nCRITICAL ERROR, [flyby_id]: '{0}' not in avaliable ids options '{1}'".format(flyby_id, avaliable_flyby_id))
			exit()

	if flyby_observation_num is not None:
		if type(flyby_observation_num) != str:
			logger.critical("\nCRITICAL ERROR, [flyby_observation_num]: Must be a str, current type = '{0}'".format(type(flyby_observation_num)))
			exit()
		if flyby_observation_num not in avaliable_observation_numbers:
			logger.critical("\nCRITICAL ERROR, [flyby_observation_num]: '{0}' not in avaliable observation options '{1}'".format(flyby_observation_num, avaliable_observation_numbers))
			exit()

	segement_options = ['S01', 'S02', 'S03']
	if segment_num is None:
		logger.critical("\nCRITICAL ERROR, [segment_num]: segment_num number required out of avaliable options {0}, none given".format(segement_options))
		exit()
	if type(segment_num) != str:
		logger.critical("\nCRITICAL ERROR, [segment_num]: Must be a str, current type = '{0}'".format(type(segment_num)))
		exit()
	if segment_num not in segement_options:
		logger.critical("\nCRITICAL ERROR, [segment_num]: '{0}' not an avaliable segment option '{1}'".format(segment_num, segement_options))
		exit()

	if len(additional_data_types_to_download) != 0:
		if type(additional_data_types_to_download) != list:
			logger.critical("\nCRITICAL ERROR [additional_data_types_to_download]: Must be a list, current type = '{0}'".format(type(additional_data_types_to_download)))
			exit()
		for data_type in additional_data_types_to_download:
			if type(data_type) != str:
				logger.critical("\nCRITICAL ERROR [additional_data_types_to_download]: All data types should be strings, but '{0}' current type = '{0}'".format(data_type, type(data_type)))

		# Get data types for the coradr type from coradr_jpl_options.csv
		coradr_data_types_available = os.path.join(os.path.dirname(__file__), 'data', 'coradr_jpl_options.csv')  # get file's directory, up one level, /data/*.csv
		df = pd.read_csv(coradr_data_types_available)
		if flyby_id is not None: 
			flyby_observation_num = pydar.convertFlybyIDToObservationNumber(flyby_id)
		coradr_versions = df[df['CORADR ID'].str.contains(flyby_observation_num)]
		version_id = ""
		if len(coradr_versions) > 1:
			for i in range(len(coradr_versions)):
				version_id = "_V0{0}".format(i+1)
		coradr_id = "CORADR_{0}{1}".format(flyby_observation_num, version_id)
		coradr_row = df.loc[df['CORADR ID'] == coradr_id]
		coradr_data_types = []
		for index, row in coradr_row.iterrows():
			row = row.tolist()
		for i, row_bool in enumerate(row[2:]): # Ignores first two columns: CORADR ID and Is Titan Flyby
			if row_bool is True:
				coradr_data_types.append(pydar.datafile_types_columns[i])

		for data_type in additional_data_types_to_download:
			if data_type not in coradr_data_types:
				logger.critical("\nCRITICAL ERROR [additional_data_types_to_download]: Data type '{0}' not avaliable in {1}".format(data_type, coradr_data_types))
				exit()

	if resolution is not None and top_x_resolutions is not None:
		logger.critical("\nCRITICAL ERROR: Requires either a resolution OR a top_x_resolutions, not both".format(type(resolution)))
		exit()

	if resolution is not None :
		if type(resolution) != str:
			logger.critical("\nCRITICAL ERROR, [resolution]: Must be a str, current type = '{0}'".format(type(resolution)))
			exit()
		if resolution not in pydar.resolution_types:
			logger.critical("\nCRITICAL ERROR, [resolution]: resolution '{0}' must be a valid resolution type in {1}".format(resolution, pydar.resolution_types))
			exit()

	if top_x_resolutions is not None:
		if type(top_x_resolutions) != int:
			logger.critical("\nCRITICAL ERROR, [top_x_resolutions]: Must be a int, current type = '{0}'".format(type(top_x_resolutions)))
			exit()
		if top_x_resolutions < 1 or top_x_resolutions > 5:
			logger.critical("\nCRITICAL ERROR, [top_x_resolutions]: Must be a value from 1 to 5, not '{0}'".format(top_x_resolutions))
			exit()

def errorHandlingConvertFlybyIDToObservationNumber(flyby_id=None):
	# Error Handling for Converting a Flyby ID into an Observation Number
	if type(flyby_id) != str:
		logger.critical("\nCRITICAL ERROR, [flyby_id]: Must be a str, current type = '{0}'".format(type(flyby_id)))
		exit()

	flyby_csv_file = os.path.join(os.path.dirname(__file__), 'data', 'cassini_flyby.csv')  # get file's directory, up one level, /data/*.csv
	flyby_dataframe = pd.read_csv(flyby_csv_file)
	valid_flyby_ids = []
	flyby_id_found = False
	for index, row in flyby_dataframe.iterrows():
		valid_flyby_ids.append(row[0])
		if row[0] == flyby_id:
			flyby_id_found = True
			break
	if not flyby_id_found:
		logger.critical("\nCRITICAL ERROR, [flyby_id]: Invalid flyby_id, '{0}', choose from:\n{1}".format(flyby_id, valid_flyby_ids))
		exit()

def errorHandlingConvertObservationNumberToFlybyID(flyby_observation_num=None):
	# Error Handling for Converting an Observation Number to a Flyby ID
	if type(flyby_observation_num) != str:
		logger.critical("\nCRITICAL ERROR, [flyby_observation_num]: Must be a str, current type = '{0}'".format(type(flyby_observation_num)))
		exit()

	flyby_csv_file = os.path.join(os.path.dirname(__file__), 'data', 'cassini_flyby.csv')  # get file's directory, up one level, /data/*.csv
	flyby_dataframe = pd.read_csv(flyby_csv_file)
	valid_observation_nums = []
	observation_num_found = False
	for index, row in flyby_dataframe.iterrows():
		take_ob_num = "0" + row[1].split(" ")[1]
		valid_observation_nums.append(take_ob_num)
		if take_ob_num == flyby_observation_num:
			observation_num_found = True
			break
	if not observation_num_found:
		logger.critical("\nCRITICAL ERROR, [flyby_observation_num]: Invalid flyby_observation_num, '{0}', choose from:\n{1}".format(observation_num, valid_observation_nums))
		exit()

def errorHandlingDisplayImages(image_directory=None):
	# Error Handling for Displaying Images from an Image Directory
	if type(image_directory) != str:
		logger.critical("\nCRITICAL ERROR, [image_directory]: Must be a str, current type = '{0}'".format(type(image_directory)))
		exit()

def errorHandlingREADME(coradr_results_directory=None,
						section_to_print=None,
						print_to_console=True):
	if type(coradr_results_directory) != str:
		logger.critical("\nCRITICAL ERROR, [coradr_results_directory]: Must be a str, current type = '{0}'".format(type(coradr_results_directory)))
		exit()

	if type(section_to_print) != str:
		logger.critical("\nCRITICAL ERROR, [section_to_print]: Must be a str, current type = '{0}'".format(type(section_to_print)))
		exit()

	if type(print_to_console) != bool:
		logger.critical("\nCRITICAL ERROR, [print_to_console]: Must be a bool, current type = '{0}'".format(type(print_to_console)))
		exit()

def errorHandlingRetrieveIDSByFeature(feature_name=None):
	if type(feature_name) != str:
		logger.critical("\nCRITICAL ERROR, [feature_name]: Must be a str, current type = '{0}'".format(type(feature_name)))
		exit()

def errorHandlingRetrieveByLatitudeLongitude(latitude=None,
											longitude=None):
	if type(latitude) != float and type(latitude) != int:
		logger.critical("\nCRITICAL ERROR, [latitude]: Must be a float or int, current type = '{0}'".format(type(latitude)))
		exit()

	if type(longitude) != float and type(longitude) != int:
		logger.critical("\nCRITICAL ERROR, [longitude]: Must be a float or int, current type = '{0}'".format(type(longitude)))
		exit()

def errorHandlingRetrieveByLatitudeLongitudeRange(northernmost_latitude=None,
												southernmost_latitude=None,
												easternmost_longitude=None,
												westernmost_longitude=None):
	if type(northernmost_latitude) != float and type(northernmost_latitude) != int:
		logger.critical("\nCRITICAL ERROR, [northernmost_latitude]: Must be a float or int, current type = '{0}'".format(type(northernmost_latitude)))
		exit()

	if type(southernmost_latitude) != float and type(southernmost_latitude) != int:
		logger.critical("\nCRITICAL ERROR, [southernmost_latitude]: Must be a float or int, current type = '{0}'".format(type(southernmost_latitude)))
		exit()

	if type(easternmost_longitude) != float and type(easternmost_longitude) != int:
		logger.critical("\nCRITICAL ERROR, [easternmost_longitude]: Must be a float or int, current type = '{0}'".format(type(easternmost_longitude)))
		exit()

	if type(westernmost_longitude) != float and type(westernmost_longitude) != int:
		logger.critical("\nCRITICAL ERROR, [westernmost_longitude]: Must be a float or int, current type = '{0}'".format(type(westernmost_longitude)))
		exit()

	if northernmost_latitude < southernmost_latitude:
		logger.critical("\nCRITICAL ERROR, [latitude]: northernmost_latitude must be greater than southernmost_latitude")
		exit()

	if westernmost_longitude < easternmost_longitude:
		logger.critical("\nCRITICAL ERROR, [longitude]: westernmost_longitude must be greater than easternmost_longitude")
		exit()


def errorHandlingRetrieveIDSByTime(year=None, month=None, day=None, hour=0, minute=0, second=0, millisecond=0):
	if type(year) != float and type(year) != int:
		logger.critical("\nCRITICAL ERROR, [year]: Must be a float or int, current type = '{0}'".format(type(year)))
		exit()

	if type(month) != float and type(month) != int:
		logger.critical("\nCRITICAL ERROR, [month]: Must be a float or int, current type = '{0}'".format(type(month)))
		exit()

	if type(day) != float and type(day) != int:
		logger.critical("\nCRITICAL ERROR, [day]: Must be a float or int, current type = '{0}'".format(type(day)))
		exit()

	if type(hour) != float and type(hour) != int:
		logger.critical("\nCRITICAL ERROR, [hour]: Must be a float or int, current type = '{0}'".format(type(hour)))
		exit()

	if type(minute) != float and type(minute) != int:
		logger.critical("\nCRITICAL ERROR, [minute]: Must be a float or int, current type = '{0}'".format(type(minute)))
		exit()

	if type(second) != float and type(second) != int:
		logger.critical("\nCRITICAL ERROR, [second]: Must be a float or int, current type = '{0}'".format(type(second)))
		exit()

	if type(millisecond) != float and type(millisecond) != int:
		logger.critical("\nCRITICAL ERROR, [millisecond]: Must be a float or int, current type = '{0}'".format(type(millisecond)))
		exit()



# Read AAREADME.TXT and .LBL sections to console
import logging
import os

import pydar

## Logging set up for .INFO
logger = logging.getLogger(__name__)
logger.setLevel(logging.INFO)
stream_handler = logging.StreamHandler()
logger.addHandler(stream_handler)

def determineSectionToPrint(section_to_print=None, aareadmeOrLBL=None):
	# check which list the section_to_print is from
	if aareadmeOrLBL == "LBL":
		if section_to_print in lblreadme_general_options:
			return lblreadme_general_options
		if section_to_print in lblreadme_section_options:
			return lblreadme_section_options
	if aareadmeOrLBL == "AAREADME":
		if section_to_print in aareadme_general_options:
			return aareadme_general_options
		if section_to_print in aareadme_section_options:
			return aareadme_section_options

#######################################################################
## AAREADME.TXT
aareadme_general_options = ["PDS_VERSION_ID",
							"RECORD_TYPE",
							"INSTRUMENT_HOST_NAME",
							"INSTRUMENT_NAME",
							"PUBLICATION_DATE",
							"NOTE",
							"Volume"]
aareadme_section_options = ["Introduction",
							"Disk Format",
							"File Formats",
							"Volume Contents",
							"Recommended DVD Drives and Driver Software",
							"Errata and Disclaimer",
							"Version Status",
							"Contact Information"]

def returnAllAAREADMEOptions():
	logger.info("Line-By-Line Options: {0}".format(aareadme_general_options))
	logger.info("Section Header Options: {0}".format(aareadme_section_options))

def readAAREADME(coradr_results_directory=None, section_to_print=None, print_to_console=True):
	# Print AAREADME to console

	pydar.errorHandlingREADME(coradr_results_directory=coradr_results_directory,
							section_to_print=section_to_print,
							print_to_console=print_to_console)

	sectionList = determineSectionToPrint(section_to_print, "AAREADME")
	if sectionList is None:
		# TODO: error handling if section_to_print doesn't exist
		logger.critical("Cannot find a revelant section_to_print: Invalid '{0}'".format(section_to_print))
		exit()

	# Define position to start console print, default to 'All' if no section is specified
	if section_to_print is None:
		start_index = 0
		start_position = sectionList[start_index]
	else:
		start_index = sectionList.index(section_to_print)
		start_position = sectionList[start_index]

	# Define position to end console print, defaults to end of file if no section is specified
	if section_to_print is None:
		end_index = None
		end_position = None
	else:
		end_index = start_index + 1
		if end_index >= len(sectionList): 
			if sectionList == aareadme_general_options:
				end_position = aareadme_section_options[0] # the start of the section list is the end of the line-by-line list
			if sectionList == aareadme_section_options:
				end_position = None # display the last element in the list
		else:
			end_position = sectionList[end_index]

	logger.debug("section_to_print = {0}".format(section_to_print))
	logger.debug("start_position = {0}".format(start_position))
	logger.debug("end_position = {0}\n".format(end_position))

	output_string = ''
	with open("{0}/AAREADME.TXT".format(coradr_results_directory), "r") as readme_file:
		within_readme_section = False
		for line in readme_file.readlines():
			if start_position in line:
				if start_position != "Volume":
					within_readme_section = True
				else:
					if "Titan Flyby T" in line:
						within_readme_section = True
			if end_position is not None:
				if end_position in line:
					if end_position != "Volume":
						break
					else:
						if "Titan Flyby T" in line:
							break
			if within_readme_section:
				if 'OBJECT' not in line and 'END' not in line:
					output_string += line

	output_string = output_string.rstrip()

	if "=" in output_string and sectionList != aareadme_section_options:
		output_string = (output_string.split("=")[1]).strip() # only return the value from the line (PDS_VERSION_ID       = PDS3 -> PDS3)
	else:
		output_string = output_string.strip() # 'Volume' option display entire row
	if print_to_console: logger.info(output_string)
	return output_string
#########################################################################

#########################################################################
## .LBL README FILES
lblreadme_general_options = ["PDS_VERSION_ID",
							"DATA_SET_ID",
							"DATA_SET_NAME",
							"PRODUCER_INSTITUTION_NAME",
							"PRODUCER_ID",
							"PRODUCER_FULL_NAME",
							"PRODUCT_ID",
							"PRODUCT_VERSION_ID",
							"INSTRUMENT_HOST_NAME",
							"INSTRUMENT_HOST_ID",
							"INSTRUMENT_NAME",
							"INSTRUMENT_ID",
							"TARGET_NAME",
							"START_TIME",
							"STOP_TIME",
							"SPACECRAFT_CLOCK_START_COUNT",
							"SPACECRAFT_CLOCK_STOP_COUNT",
							"PRODUCT_CREATION_TIME",
							"SOURCE_PRODUCT_ID",
							"MISSION_PHASE_NAME",
							"MISSION_NAME",
							"SOFTWARE_VERSION_ID",
							"FILE_NAME COMPRESSED",
							"RECORD_TYPE COMPRESSED",
							"ENCODING_TYPE",
							"INTERCHANGE_FORMAT",
							"UNCOMPRESSED_FILE_NAME",
							"REQUIRED_STORAGE_BYTES",
							"^DESCRIPTION",
							"FILE_NAME UNCOMPRESSED",
							"RECORD_TYPE UNCOMPRESSED",
							"RECORD_BYTES",
							"FILE_RECORDS",
							"LABEL_RECORDS",
							"^IMAGE",
							"LINES",
							"LINE_SAMPLES",
							"SAMPLE_TYPE",
							"SAMPLE_BITS",
							"CHECKSUM",
							"SCALING_FACTOR",
							"OFFSET",
							"MISSING_CONSTANT",
							"NOTE",
							"^DATA_SET_MAP_PROJECTION",
							"MAP_PROJECTION_TYPE",
							"FIRST_STANDARD_PARALLEL",
							"SECOND_STANDARD_PARALLEL",
							"A_AXIS_RADIUS",
							"B_AXIS_RADIUS",
							"C_AXIS_RADIUS",
							"POSITIVE_LONGITUDE_DIRECTION",
							"CENTER_LATITUDE",
							"CENTER_LONGITUDE",
							"REFERENCE_LATITUDE",
							"REFERENCE_LONGITUDE",
							"LINE_FIRST_PIXEL",
							"LINE_LAST_PIXEL",
							"SAMPLE_FIRST_PIXEL",
							"SAMPLE_LAST_PIXEL",
							"MAP_PROJECTION_ROTATION",
							"MAP_RESOLUTION",
							"MAP_SCALE",
							"MAXIMUM_LATITUDE",
							"MINIMUM_LATITUDE",
							"EASTERNMOST_LONGITUDE",
							"WESTERNMOST_LONGITUDE",
							"LINE_PROJECTION_OFFSET",
							"SAMPLE_PROJECTION_OFFSET",
							"OBLIQUE_PROJ_POLE_LATITUDE",
							"OBLIQUE_PROJ_POLE_LONGITUDE",
							"OBLIQUE_PROJ_POLE_ROTATION",
							"OBLIQUE_PROJ_X_AXIS_VECTOR",
							"OBLIQUE_PROJ_Y_AXIS_VECTOR",
							"OBLIQUE_PROJ_Z_AXIS_VECTOR",
							"LOOK_DIRECTION",
							"COORDINATE_SYSTEM_NAME",
							"COORDINATE_SYSTEM_TYPE"]
lblreadme_section_options = ["PRODUCT DESCRIPTION",
							"DESCRIPTION OF COMPRESSED AND UNCOMPRESSED FILES",
							"POINTERS TO START RECORDS OF OBJECTS IN FILE",
							"DESCRIPTION OF OBJECTS CONTAINED IN FILE"]

def returnAllLBLOptions():
	# Print out all the .LBL options
	logger.info("Line-By-Line Options: {0}".format(lblreadme_general_options))
	logger.info("Section Header Options: {0}".format(lblreadme_section_options))

def readLBLREADME(coradr_results_directory=None, section_to_print=None, print_to_console=True):
	# Print .LBL to console

	if section_to_print == "FILE_NAME" or section_to_print == "RECORD_TYPE":
		logger.critical("CRITICAL ERROR: Specify {0} as either '{0} UNCOMPRESSED' or '{0} COMPRESSED'".format(section_to_print))
		exit()
	# Catch common mispelling: not including the ^ at the front of a line name
	if section_to_print == "DESCRIPTION" or section_to_print == "IMAGE" or section_to_print == "DATA_SET_MAP_PROJECTION":
		section_to_print = "^{0}".format(section_to_print) # set to specific object, easy to miss the ^

	pydar.errorHandlingREADME(coradr_results_directory=coradr_results_directory,
							section_to_print=section_to_print,
							print_to_console=print_to_console)

	sectionList = determineSectionToPrint(section_to_print, "LBL")
	if sectionList is None:
		# TODO: error handling if section_to_print doesn't exist
		logger.critical("Cannot find a revelant section_to_print: Invalid '{0}'".format(section_to_print))
		exit()

	# Define position to start console print, default to 'All' if no section is specified
	if section_to_print is None:
		start_index = 0
		start_position = sectionList[start_index]
	else:
		start_index = sectionList.index(section_to_print)
		start_position = sectionList[start_index]
	if "FILE_NAME " in section_to_print or "RECORD_TYPE " in section_to_print:
		# two repeated types, get the files before the repeated option
		if "COMPRESSED" in section_to_print:
			start_index = sectionList.index("SOFTWARE_VERSION_ID")
			start_position = sectionList[start_index]
		if "UNCOMPRESSED" in section_to_print:
			start_index = sectionList.index("REQUIRED_STORAGE_BYTES")
			start_position = sectionList[start_index]

	# Define position to end console print, defaults to end of file if no section is specified
	if section_to_print is None:
		end_index = None
		end_position = None
	else:
		end_index = start_index + 1
		if end_index >= len(sectionList): 
			end_position = None # display the last element in the list
		else:
			end_position = sectionList[end_index]
	if "FILE_NAME " in section_to_print or "RECORD_TYPE " in section_to_print:
		# two repeated types, get the files before the repeated option
		if "COMPRESSED" in section_to_print:
			end_index = sectionList.index("INTERCHANGE_FORMAT")
			end_position = sectionList[end_index]
		if "UNCOMPRESSED" in section_to_print:
			end_index = sectionList.index("FILE_RECORDS")
			end_position = sectionList[end_index]

	output_string = ''
	lbl_file = [i for i in os.listdir(coradr_results_directory) if '.LBL' in i]
	if len(lbl_file) != 1:
		if len(lbl_file) == 0:
			logger.critical("No .LBL file found at {0}".format(coradr_results_directory))
			exit()
		if len(lbl_file) > 1:
			logger.critical("Multiple .LBL file found = {0}, need to choose one to read from".format(lbl_file))
			exit() # TODO: error handling to check that .LBL exists
	lbl_file = lbl_file[0]

	logger.debug("section_to_print = {0}".format(section_to_print))
	logger.debug("start_position = {0}".format(start_position))
	logger.debug("end_position = {0}\n".format(end_position))

	output_string = ''
	with open("{0}/{1}".format(coradr_results_directory, lbl_file), "r") as readme_file:
		within_readme_section = False
		for line in readme_file.readlines():
			if start_position in line:
				within_readme_section = True
			if end_position is not None:
				if end_position in line:
					break
			if within_readme_section:
				if sectionList is not lblreadme_section_options:
					if "/*" not in line: # if checking individual values, ignore /**/ section headers
						output_string += line
					else:
						if "FILE_NAME " not in section_to_print and "RECORD_TYPE " not in section_to_print:
							break
				else:
					output_string += line # collection by section

	repeated_or_skipped_values_to_find = ["FILE_NAME", "RECORD_TYPE", "^DESCRIPTION", "COORDINATE_SYSTEM_TYPE"]
	output_lines = output_string.split("\n")
	if section_to_print.split(" ")[0] in repeated_or_skipped_values_to_find:
		for line in output_lines:
			if section_to_print.split(" ")[0] in line:
				output_string = line
	
	if section_to_print == "NOTE":
		output_string = ""
		for line in output_lines:
			if "END_OBJECT" not in line:
				output_string += line
			else:
				break

	output_string = output_string.rstrip()
	if sectionList != lblreadme_section_options:
		output_string = (output_string.split("=")[1]).strip() # only return the value from the line (PDS_VERSION_ID       = PDS3 -> PDS3)
	if print_to_console: logger.info(output_string)
	return output_string

########################################################################



# Extract metadata from SBDR files (.TAB files)
import numpy as np
import pdr

def extractMetadata():
	# Note: need both the .TAB and the .FMT file to run
	
	tab_file = "pydar/testing_files/SBDR_15_D065_V03.TAB"
	SBDR_FILE = pdr.read(tab_file)
	#print(SBDR_FILE['SBDR_TABLE'])
	# Each row is Burst Data
	#print("Headers = {0}".format(list(SBDR_FILE['SBDR_TABLE'])))

	# TODO: Associated Burt Data with Image Data
	# RADAR_MODE:
	# 	0 or 8 = Scatterometry
	# 	1 or 9 = Altimetry
	#	2 or 10 = Low Res SAR
	# 	3 or 11= High Res SAR
	# 	4 or 12 = Radiometry
	# 	where +8 = Auto-Gain enabled
	radar_mode = SBDR_FILE['SBDR_TABLE']['RADAR_MODE']
	sbdr = SBDR_FILE['SBDR_TABLE']

	#print(radar_mode)
	
	# identify the low and high res sar with and without autogain
	low_res_sar_burst = sbdr.index[sbdr.RADAR_MODE == 3].tolist()
	if len(sbdr.index[sbdr.RADAR_MODE == 11].tolist()) > 0:
		low_res_sar_burst = low_res_sar_burst + sbdr.index[sbdr.RADAR_MODE == 11].tolist()

	hi_res_sar_burst = sbdr.index[sbdr.RADAR_MODE == 4].tolist()
	if len(sbdr.index[sbdr.RADAR_MODE == 12].tolist()) > 0:
		hi_res_sar_burst = hi_res_sar_burst + sbdr.index[sbdr.RADAR_MODE == 12].tolist()

	# collect sar data indices into one list
	sar_ind = low_res_sar_burst + hi_res_sar_burst
	sbdr_sar = sbdr.iloc[sar_ind]
	print('Unique Beam Patterns before filtering: '+str(sbdr_sar['BEM'].unique()))

	# filter SAR data for best active points (Alex's #97)
	# Active Point: active or passive sensor, but only gets when radar when in an active state
	# ACT_AZIMUTH_ANGLE: Direction of the projection of the antenna look vector into the plane 
	#					 tangent the surface at the center of the measurement as an angle CCW from East 
	#					 (such that N is 90 deg)
	# ACT_ELLIPSE_PT1_LAT: Latitude of the first point (on major axis) in the ellipse marking the 
	# 					   active measurement two way 3-dB gain pattern. 
	sbdr_sar = sbdr_sar[sbdr_sar['ACT_AZIMUTH_ANGLE'] != 0]
	sbdr_sar = sbdr_sar[sbdr_sar['ACT_ELLIPSE_PT1_LAT'] != 0]
	
	print('Found {0} active beam pulses in SAR'.format(len(sbdr_sar)))
	# total width of the RADAR swath is created by combining the five individually illuminated subbeams 
	# each bursts use different beam(s) while taking SAR measurements
	beam_1 = [] # DEFINE: Smallest look angle subbeam 
	beam_2 = [] # DEFINE: Second-smallest look angle subbeam
	beam_3 = [] # DEFINE: Middle subbeam swath with greatest gain
	beam_4 = [] # DEFINE: Second-largest look angle subbeam
	beam_5 = [] # DEFINE: Largest look angle subbeam

	print('Unique Beam Patterns used after filtering: '+str(sbdr_sar['BEM'].unique()))
		
		
	for x in sbdr_sar['BEM']:
		binx = bin(x)[2:].zfill(2)
		beam_5.append(binx[0])
		beam_4.append(binx[1])
		beam_3.append(binx[2])
		beam_2.append(binx[3])
		beam_1.append(binx[4])
		
	
	#beam = sbdr_sar['BEM'].tolist()
	#print('size of beam',+len(sbdr_sar['BEM']))
	# convert to integer and sum
	beam_1 = list(map(int, beam_1))
	beam_2 = list(map(int, beam_2))
	beam_3 = list(map(int, beam_3))
	beam_4 = list(map(int, beam_4))
	beam_5 = list(map(int, beam_5))

	print("# of Bursts with Beam 1: "+str(sum(beam_1)))
	print("# of Bursts with Beam 2: "+str(sum(beam_2)))
	print("# of Bursts with Beam 3: "+str(sum(beam_3)))
	print("# of Bursts with Beam 4: "+str(sum(beam_4)))
	print("# of Bursts with Beam 5: "+str(sum(beam_5)))
	print("# of Bursts Total: "+str(len(sbdr_sar['BEM'])))

'''
	#img_file = "pydar_results/CORADR_0211_V03_S01/BIBQD78S004_D211_T065S01_V03.IMG"
	#BIDR_FILE = pdr.read(img_file)
	#print(BIDR_FILE.keys())

	#['.start_burst_num'] and ['.end_burst_num']
'''



# -*- coding: utf-8 -*-

# Python Package Setup
from setuptools import setup, find_namespace_packages

VERSION="0.1.0"
DESCRIPTION="A Python package to access, download, view, and manipulate Cassini RADAR images"

with open("README.md", "r") as f:
	long_description_readme = f.read()

setup(
	name="pydar",
	version=VERSION,
	description=DESCRIPTION,
	long_description=long_description_readme,
	long_description_content_type='text/markdown',
	url="https://github.com/unaschneck/pydar",
	download_url="https://github.com/unaschneck/pydar/archive/refs/tags/v{0}.tar.gz".format(VERSION),
	author="Una Schneck (unaschneck), C. Y. Schneck (cyschneck)",
	keywords=[],
	license="MIT",
	classifiers=[
		"Development Status :: 1 - Planning",
		"Intended Audience :: Developers",
		"Intended Audience :: Education",
		"Intended Audience :: Science/Research",
		"License :: OSI Approved :: MIT License",
		"Programming Language :: Python",
		"Programming Language :: Python :: 3.9",
		"Intended Audience :: Education",
		"Intended Audience :: Science/Research",
		"Topic :: Scientific/Engineering :: Physics",
		"Topic :: Scientific/Engineering :: Visualization",
		"Topic :: Scientific/Engineering :: Astronomy"
	],
	packages=find_namespace_packages(include=['pydar', 'pydar.*']),
	include_package_data=True,
	install_requires=[
			"beautifulsoup4>=4.11.1",
			"matplotlib>=3.1.0",
			"pdf>=0.7.2",
			"planetaryimage>=0.5.0",
			"urllib3>=1.26.9"
			],
	python_requires='>=3.9'
)



# Retrieve Flyby Observation and IDs based on Feature Name, Latitude/Longitude or Time
import logging
import math
import os

import pandas as pd
import numpy as np

import pydar

## Logging set up for .INFO
logger = logging.getLogger(__name__)
logger.setLevel(logging.DEBUG)
stream_handler = logging.StreamHandler()
logger.addHandler(stream_handler)

def latitudeLongitudeWithFeatureNameFromCSV():
	# Retrieve a list of Feature Names with a range of the associated latitude/longitude values
	feature_name_dict = {}

	flyby_csv_file = os.path.join(os.path.dirname(__file__), 'data', 'feature_name_details.csv')  # get file's directory, up one level, /data/*.csv
	flyby_dataframe = pd.read_csv(flyby_csv_file)

	for index, row in flyby_dataframe.iterrows():
			feature_name_dict[row["Feature Name"]] = {"Southernmost Latitude": row["Southernmost Latitude"],
													"Northernmost Latitude": row["Northernmost Latitude"], 
													"Easternmost Longitude": row["Easternmost Longitude"], 
													"Westernmost Longitude": row["Westernmost Longitude"],
													"Center Latitude": row["Center Latitude"],
													"Center Longitude": row["Center Longitude"]}

	return feature_name_dict

def retrieveIDSByFeatureName(feature_name=None):
	pydar.errorHandlingRetrieveIDSByFeature(feature_name=feature_name)

	feature_name_csv_dict = latitudeLongitudeWithFeatureNameFromCSV()
	feature_name = feature_name.title() # convert 'ligeria mare' to 'Ligeria Mare' to make input not sensitive to case

	if feature_name not in feature_name_csv_dict.keys():
		logger.critical("Feature Name '{0}' not in available in features list = {1}".format(feature_name, list(feature_name_csv_dict.keys())))
		exit()

	feature_dict = feature_name_csv_dict[feature_name]
	flyby_ids = retrieveIDSByLatitudeLongitudeRange(northernmost_latitude=feature_dict["Northernmost Latitude"],
													southernmost_latitude=feature_dict["Southernmost Latitude"],
													easternmost_longitude=feature_dict["Easternmost Longitude"],
													westernmost_longitude=feature_dict["Westernmost Longitude"])
	return flyby_ids

def retrieveIDSByLatitudeLongitude(latitude=None, longitude=None):
	# Retrieve all FLyby Ids at a specific latitude/longitude
	pydar.errorHandlingRetrieveByLatitudeLongitude(latitude=latitude, longitude=longitude)

	# Runs range check, but the range is 0 for an exact spot
	flyby_ids = retrieveIDSByLatitudeLongitudeRange(northernmost_latitude=latitude,
													southernmost_latitude=latitude,
													easternmost_longitude=longitude,
													westernmost_longitude=longitude)
	return flyby_ids

def retrieveIDSByLatitudeLongitudeRange(northernmost_latitude=None,
										southernmost_latitude=None,
										easternmost_longitude=None,
										westernmost_longitude=None):
	# Retrieve all Flyby Ids that cover a specific latitude/longitude or within a range of latitude/longitudes
	pydar.errorHandlingRetrieveByLatitudeLongitudeRange(northernmost_latitude=northernmost_latitude,
														southernmost_latitude=southernmost_latitude,
														easternmost_longitude=easternmost_longitude,
														westernmost_longitude=westernmost_longitude)

	swath_csv_file = os.path.join(os.path.dirname(__file__), 'data', 'swath_coverage_by_time_position.csv')  # get file's directory, up one level, /data/*.csv
	swath_dataframe = pd.read_csv(swath_csv_file)

	flyby_ids = {} # {'flyby_id': ['seg1', seg4']
	for index, row in swath_dataframe.iterrows():
		flyby = str(row['FLYBY ID'])
		if float(row["MINIMUM_LATITUDE (Degrees)"]) <= southernmost_latitude and float(row["MAXIMUM_LATITUDE (Degrees)"]) >= northernmost_latitude:
			if float(row["EASTERNMOST_LONGITUDE (Degrees)"]) <= easternmost_longitude and float(row["WESTERNMOST_LONGITUDE (Degrees)"]) >= westernmost_longitude:
				if flyby not in flyby_ids.keys():
					flyby_ids[flyby] = []
				segment_number = "S0" + str(row["SEGMENT NUMBER"])
				if segment_number not in flyby_ids[flyby]:
					flyby_ids[flyby].append(segment_number)

	if len(flyby_ids) == 0:
		logger.info("\n[WARNING]: No flyby IDs found at latitude from {0} N to {1} S and longitude from {2} W to {3} E\n".format(northernmost_latitude,
																																southernmost_latitude,
																																westernmost_longitude,
																																easternmost_longitude))
		exit()

	return flyby_ids

def retrieveIDSByTime(year=None, month=None, day=None, hour=0, minute=0, second=0, millisecond=0):
	# TODO: Retrieve Flyby IDs based on a Timestamp
	pydar.errorHandlingRetrieveIDSByTime(year=year, month=month, day=day, hour=hour, minute=minute, second=second, millisecond=millisecond)

	logger.info("TODO: {0} year, {1} month, {2} day, {3} hour, {4} minute, {5} second, {6} millisecond".format(year,
																											month,
																											day,
																											hour,
																											minute,
																											second,
																											millisecond))

	swath_csv_file = os.path.join(os.path.dirname(__file__), 'data', 'swath_coverage_by_time_position.csv')  # get file's directory, up one level, /data/*.csv
	swath_dataframe = pd.read_csv(swath_csv_file)

	flyby_ids = {} # {'flyby_id': ['seg1', seg4']
	for index, row in swath_dataframe.iterrows():
		flyby = str(row['FLYBY ID'])
		'''
		if float(row["MINIMUM_LATITUDE (Degrees)"]) <= southernmost_latitude and float(row["MAXIMUM_LATITUDE (Degrees)"]) >= northernmost_latitude:
			if flyby not in flyby_ids.keys():
				flyby_ids[flyby] = []
			segment_number = "S0" + str(row["SEGMENT NUMBER"])
			if segment_number not in flyby_ids[flyby]:
				flyby_ids[flyby].append(segment_number)
		'''

	return flyby_ids

def retrieveFeaturesFromLatitudeLongitude(latitude=None, longitude=None):
	# Retrieve all Feature Names that at a specific latitude/longitude
	pydar.errorHandlingRetrieveByLatitudeLongitude(latitude=latitude, longitude=longitude)

	# Runs range check, but the range is 0 for an exact spot
	feature_names_list = retrieveFeaturesFromLatitudeLongitudeRange(northernmost_latitude=latitude,
																	southernmost_latitude=latitude,
																	easternmost_longitude=longitude,
																	westernmost_longitude=longitude)
	return feature_names_list


def retrieveFeaturesFromLatitudeLongitudeRange(northernmost_latitude=None,
												southernmost_latitude=None,
												easternmost_longitude=None,
												westernmost_longitude=None):
	# Retrieve all Feature Names that are within a range of latitude/longitude
	pydar.errorHandlingRetrieveByLatitudeLongitudeRange(northernmost_latitude=northernmost_latitude,
														southernmost_latitude=southernmost_latitude,
														easternmost_longitude=easternmost_longitude,
														westernmost_longitude=westernmost_longitude)

	feature_name_csv_dict = latitudeLongitudeWithFeatureNameFromCSV()
	feature_names_list = []

	for feature_name, position_dict in feature_name_csv_dict.items():
		if float(position_dict["Southernmost Latitude"]) >= southernmost_latitude and float(position_dict["Northernmost Latitude"]) <= northernmost_latitude:
			if float(position_dict["Easternmost Longitude"]) >= easternmost_longitude and float(position_dict["Westernmost Longitude"]) <= westernmost_longitude:
				feature_names_list.append(feature_name)

	if len(feature_names_list) == 0:
		logger.info("\n[WARNING]: No Features found at latitude from {0} N to {1} S and longitude from {2} W to {3} E\n".format(northernmost_latitude,
																																southernmost_latitude,
																																westernmost_longitude,
																																easternmost_longitude))
		exit()

	return feature_names_list
